/*
 *  Copyright (c) 2020 NetApp
 *  All rights reserved
 */

package nvc

import (
	"context"
	"fmt"
	"net/http"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/util"

	//Quark
	corev1 "k8s.io/api/core/v1"
	"quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

var srv *http.Server

var (
	provisonedProtocol          = "netapp_volume_controller_provisioned_protocol_count"
	provisionedHelp             = "Total number of volumes based on protocol"
	provisionedLabels  []string = []string{"protocol", "storagePoolUUID"}
	desc               *prometheus.Desc

	KerberosEnabledCount = "netapp_volume_controller_kerberos_vol_count"
	KerberosEnableHelp   = "Total number of volume having kerberos enabled"
	kerbLabels           = []string{"storagePoolUUID"}
	kerbDesc             *prometheus.Desc

	spFeatureSettingValueMetricName          = "netapp_volume_controller_sp_feature_setting_value"
	spFeatureSettingValueMetricHelp          = "Displays the spec value(enabled=1/disabled=0) of feature flags in a storage pool"
	spFeatureSettingValueLabels     []string = []string{"featureName", "sdeSPUUID", "storagePoolUUID"}
	spFeatureSettingValueDesc       *prometheus.Desc

	spFeatureSettingReconciledMetricName          = "netapp_volume_controller_sp_feature_setting_reconciled"
	spFeatureSettingReconciledMetricHelp          = "Displays the reconciled status of feature flags in a storage pool. It is set to true if the status matches the desired state(spec) of the feature flag."
	spFeatureSettingReconciledLabels     []string = []string{"featureName", "sdeSPUUID", "storagePoolUUID"}
	spFeatureSettingReconciledDesc       *prometheus.Desc

	spFeatureWarningMetricName          = "netapp_volume_controller_sp_feature_warning"
	spFeatureWarningMetricHelp          = "Set to 1 if a feature flag validation error is present, meaning features could not be disabled in the storage pool due to their validation type (\"nonreversible/one-way\")"
	spFeatureWarninggLabels    []string = []string{"sdeSPUUID", "storagePoolUUID"}
	spFeatureWarningDesc       *prometheus.Desc

	quarkFeatureMetricValueName          = "netapp_volume_controller_quark_feature_setting_value"
	quarkFeatureMetricValueHelp          = "Displays the spec value(enabled=1/disabled=0) of feature flags in the quark cr"
	quarkFeatureValueLabels     []string = []string{"featureName", "quarkName"}
	quarkFeatureValueDesc       *prometheus.Desc

	quarkFeatureReconciledMetricName          = "netapp_volume_controller_quark_feature_setting_reconciled"
	quarkFeatureReconciledMetricHelp          = "Displays the reconciled status of feature flags in the quark cr. It is set to true if the status matches the desired state(spec) of the feature flag."
	quarkFeatureReconciledLabels     []string = []string{"featureName", "quarkName"}
	quarkFeatureReconciledDesc       *prometheus.Desc
)

func (c *Controller) Describe(ch chan<- *prometheus.Desc) {
	// Describe new metric
	desc = prometheus.NewDesc(provisonedProtocol, provisionedHelp, provisionedLabels, nil)
	ch <- desc

	kerbDesc = prometheus.NewDesc(KerberosEnabledCount, KerberosEnableHelp, kerbLabels, nil)
	ch <- kerbDesc

	spFeatureSettingValueDesc = prometheus.NewDesc(spFeatureSettingValueMetricName, spFeatureSettingValueMetricHelp, spFeatureSettingValueLabels, nil)
	ch <- spFeatureSettingValueDesc

	spFeatureSettingReconciledDesc = prometheus.NewDesc(spFeatureSettingReconciledMetricName, spFeatureSettingReconciledMetricHelp, spFeatureSettingReconciledLabels, nil)
	ch <- spFeatureSettingReconciledDesc

	spFeatureWarningDesc = prometheus.NewDesc(spFeatureWarningMetricName, spFeatureWarningMetricHelp, spFeatureWarninggLabels, nil)
	ch <- spFeatureWarningDesc

	quarkFeatureValueDesc = prometheus.NewDesc(quarkFeatureMetricValueName, quarkFeatureMetricValueHelp, quarkFeatureValueLabels, nil)
	ch <- quarkFeatureValueDesc

	quarkFeatureReconciledDesc = prometheus.NewDesc(quarkFeatureReconciledMetricName, quarkFeatureReconciledMetricHelp, quarkFeatureReconciledLabels, nil)
	ch <- quarkFeatureReconciledDesc
}

func (c *Controller) Collect(ch chan<- prometheus.Metric) {

	kerbCount := 0

	// List existing storage pools
	sPools, err := c.storagePoolLister.List(labels.NewSelector())
	if err != nil {
		fmt.Errorf("error listing storage pool %v", err)
	}
	// For each storage pool, list volumes related to it.
	// and calculate number of nfs3 and cifs volume
	for _, sp := range sPools {

		// If present, get the SDE(specific) SP UUID from the storage pool labels
		sdeSpUUID := ""
		if sp.Labels != nil {
			sdeSpUUID = sp.Labels[quarkconfig.SDESpUUID]
		}

		nfs3, cifs, nfs4, dual := 0, 0, 0, 0
		ls, _ := labels.Parse("quark.netapp.io/storagePool=" + sp.Name)

		vols, err := c.volumesLister.List(ls)
		if err != nil {
			fmt.Errorf("error listing volumes %v", err)
			return
		}
		for _, vol := range vols {

			if isOnline(vol) {
				switch vol.Spec.Protocol {
				case "nfs3":
					nfs3 = nfs3 + 1
				case "cifs":
					cifs = cifs + 1
				case "nfs4.1":
					nfs4 = nfs4 + 1
				case "nfs4.1,nfs3":
					dual = dual + 1
				case "nfs3,nfs4.1":
					dual = dual + 1
				}
			}
			if vol.Spec.KerberosEnabled {
				kerbCount = kerbCount + 1
			}

		}
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(nfs3),
			[]string{"nfs3", sp.Name}...,
		)
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(cifs),
			[]string{"cifs", sp.Name}...,
		)
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(nfs4),
			[]string{"nfs4", sp.Name}...,
		)
		ch <- prometheus.MustNewConstMetric(
			desc,
			prometheus.GaugeValue,
			float64(dual),
			[]string{"nfs4,nfs3.1", sp.Name}...,
		)
		ch <- prometheus.MustNewConstMetric(
			kerbDesc,
			prometheus.GaugeValue,
			float64(kerbCount),
			[]string{sp.Name}...,
		)

		// Create metrics for feature flags in a SP
		if sp.Spec.FeatureFlags != nil {
			for fName, fValue := range sp.Spec.FeatureFlags {
				featureNameStr := fmt.Sprintf("%s", fName)
				featureEnabled := float64(0)
				featureReconciled := float64(0)
				if sp.Status.FeatureFlags != nil && fValue == sp.Status.FeatureFlags[fName] {
					featureReconciled = 1
				}
				if fValue == v1alpha1.Enabled {
					featureEnabled = 1
				}
				ch <- prometheus.MustNewConstMetric(
					spFeatureSettingValueDesc,
					prometheus.GaugeValue,
					featureEnabled,
					// values for fields defined in spFeatureSettingLabels
					[]string{featureNameStr, sdeSpUUID, sp.Name}...,
				)

				ch <- prometheus.MustNewConstMetric(
					spFeatureSettingReconciledDesc,
					prometheus.GaugeValue,
					featureReconciled,
					// values for fields defined in spFeatureSettingLabels
					[]string{featureNameStr, sdeSpUUID, sp.Name}...,
				)
			}
		}

		// Create metrics for feature flags warning conditions in a SP
		warnCond := util.GetNetAppStoragePoolCondition(sp, v1alpha1.NetAppStoragePoolFeatureFlagModifyError)
		warningExists := float64(0)
		if warnCond != nil && warnCond.Status == corev1.ConditionTrue {
			warningExists = 1
		}
		ch <- prometheus.MustNewConstMetric(
			spFeatureWarningDesc,
			prometheus.GaugeValue,
			warningExists,
			// values for fields defined in spFeatureWarningLabels
			[]string{sdeSpUUID, sp.Name}...,
		)
	}

	// Quark feature flag metrics
	quarks, err := c.QuarkClientset.QuarkV1alpha1().Quarks(c.Namespace).List(context.TODO(), metav1.ListOptions{})
	if err != nil {
		log.Warn("Error listing quarks, skipping quark feature flag metrics")
	} else {
		// Produce feature metrics for each Quark (from any NS, there should only be a single instance)
		for _, quark := range quarks.Items {
			// Ensure the Quark has feature flags defined to avoid nil pointer dereference
			if quark.Status.FeatureFlags != nil {
				for fName, fValue := range quark.Spec.FeatureFlags {
					featureNameStr := fmt.Sprintf("%s", fName)
					featureEnabled := float64(0)
					if fValue.State == v1alpha1.Enabled {
						featureEnabled = 1
					}
					featureReconciled := float64(0)
					if quark.Spec.FeatureFlags != nil && fValue.State == quark.Status.FeatureFlags[fName].State {
						featureReconciled = 1
					}
					ch <- prometheus.MustNewConstMetric(
						quarkFeatureValueDesc,
						prometheus.GaugeValue,
						featureEnabled,
						//values for fields defined in quarkFeatureLabels
						[]string{featureNameStr, quark.Name}...,
					)
					ch <- prometheus.MustNewConstMetric(
						quarkFeatureReconciledDesc,
						prometheus.GaugeValue,
						featureReconciled,
						//values for fields defined in quarkFeatureLabels
						[]string{featureNameStr, quark.Name}...,
					)
				}
			}
		}
	}
}

var (
	// Provision time for last volume by NVC
	volumeProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provision_time",
			Help:      "Total time spent provisioning last Netapp Volume.",
		},
	)

	averageProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_avg_provision_time",
			Help:      "Average time to provision Netapp Volumes.",
		},
	)

	provisionedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_provisioned_count_total",
			Help:      "Total count of successfully provisioned Netapp Volumes.",
		},
	)

	failedVolumesCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failed_volume_count_total",
			Help:      "Total count of Netapp Volumes failed to provision.",
		},
	)

	knownVolumesCount = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_known_volume_count_total",
			Help:      "Total count of Netapp Volumes known to Netapp Volume Controller.",
		},
	)

	failureCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapvol_failure_count_total",
			Help:      "Total difference in actual state and desired state of volumes for Netapp Volume Controller.",
		},
	)

	provisionedStoragePoolCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_provisioned_count_total",
			Help:      "Total count of successfully provisioned Netapp Storage Pools.",
		},
	)

	failedStoragePoolsCount = prometheus.NewCounter(
		prometheus.CounterOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_failed_storage_pool_count_total",
			Help:      "Total count of Netapp Storage Pools failed to provision.",
		},
	)

	knownStoragePoolsCount = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_known_storage_pool_count_total",
			Help:      "Total count of Netapp Storage Pools known to Netapp Volume Controller.",
		},
	)

	storagePoolProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_provision_time",
			Help:      "Total time spent provisioning last Netapp Storage Pool.",
		},
	)

	storagePoolAverageProvisionTimeSeconds = prometheus.NewGauge(
		prometheus.GaugeOpts{
			Subsystem: "netapp_volume_controller",
			Name:      "ntapsp_avg_provision_time",
			Help:      "Average time to provision Netapp Storage Pools.",
		},
	)
)

func updateVolumeProvisionTimeSeconds(provisionTime int) {
	volumeProvisionTimeSeconds.Set(float64(provisionTime))
}

func updateAverageProvisionTimeSeconds(avgTime int) {
	averageProvisionTimeSeconds.Set(float64(avgTime))
}

func incFailedVolumesCount() {
	failedVolumesCount.Inc()
}

func incFailureCount() {
	failureCount.Inc()
}

func updateKnownVolumesCount(knownVolCount int) {
	knownVolumesCount.Set(float64(knownVolCount))
}

func incProvisionedVolumesCount() {
	provisionedVolumesCount.Inc()
}

func incFailedStoragePoolCount() {
	failedStoragePoolsCount.Inc()
}

func updateKnownStoragePoolCount(knownCount int) {
	knownStoragePoolsCount.Set(float64(knownCount))
}

func incProvisionedStoragePoolsCount() {
	provisionedStoragePoolCount.Inc()
}

func updateStoragePoolProvisionTimeSeconds(provisionTime int) {
	storagePoolProvisionTimeSeconds.Set(float64(provisionTime))
}

func updateStoragePoolAverageProvisionTimeSeconds(avgTime int) {
	storagePoolAverageProvisionTimeSeconds.Set(float64(avgTime))
}

func isOnline(volume *v1alpha1.NetAppVolume) bool {
	for _, cond := range volume.Status.Conditions {
		if cond.Type == v1alpha1.NetAppVolumeOnline && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}

// StartMetricsServer will start metrics exporter for nvc
func StartMetricsServer() {

	prometheus.MustRegister(volumeProvisionTimeSeconds)
	prometheus.MustRegister(averageProvisionTimeSeconds)
	prometheus.MustRegister(provisionedVolumesCount)
	prometheus.MustRegister(failedVolumesCount)
	prometheus.MustRegister(knownVolumesCount)
	prometheus.MustRegister(failureCount)
	prometheus.MustRegister(provisionedStoragePoolCount)
	prometheus.MustRegister(failedStoragePoolsCount)
	prometheus.MustRegister(knownStoragePoolsCount)
	prometheus.MustRegister(storagePoolProvisionTimeSeconds)
	prometheus.MustRegister(storagePoolAverageProvisionTimeSeconds)

	server := http.NewServeMux()
	// Handle the endpoint serving the metrics
	server.Handle("/metrics", promhttp.Handler())
	// Create a server object
	srv = &http.Server{
		Addr:    ":" + quarkconfig.NvcMetricsPort,
		Handler: server,
	}

	go func() {
		if err := srv.ListenAndServe(); err != nil && err != http.ErrServerClosed {
			log.Fatalf("Error starting nvc exporter: %s\n", err)
		}
	}()
	log.Info("NVC exporter started")
}

// StopMetricsServer will stop metrics exporter when stop signal is received
func StopMetricsServer() {
	log.Info("NVC exporter shuting down")
	ctx, _ := context.WithTimeout(context.Background(), 5*time.Second)
	if err := srv.Shutdown(ctx); err != nil {
		log.Fatalf("NVC exporter shutdown Failed: %+v", err)
	}
}
