package v2

import (
	"context"
	"fmt"
	"strings"

	acnv1alpha1 "github.com/Azure/azure-container-networking/crd/multitenancy/api/v1alpha1"
	"github.com/google/uuid"

	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	ctrlClient "sigs.k8s.io/controller-runtime/pkg/client"

	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// ConstructAzurePodNetwork creates a PodNetwork resource
func ConstructAzurePodNetwork(
	name string,
	subnetGUID string,
	subnetResourceID string,
	vnetGUID string,
	salOverrideToken string,
) (podNetwork *acnv1alpha1.PodNetwork, err error) {

	// Validate name
	name = strings.TrimSpace(name)
	if len(name) == 0 {
		return nil, fmt.Errorf("name cannot be empty string")
	}

	// Validate subnetGUID
	if _, err = uuid.Parse(subnetGUID); err != nil {
		return nil, fmt.Errorf("subnetGUID is not a valid UUID: %v", err)
	}

	// Validate vnetGUID
	if _, err = uuid.Parse(vnetGUID); err != nil {
		return nil, fmt.Errorf("vnetGUID is not a valid UUID: %v", err)
	}

	// Validate subnetResourceID
	subnetResourceID = strings.TrimSpace(subnetResourceID)
	if len(subnetResourceID) == 0 {
		return nil, fmt.Errorf("subnetResourceID cannot be empty string")
	}

	podNetwork = &acnv1alpha1.PodNetwork{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: name,
		},
		Spec: acnv1alpha1.PodNetworkSpec{
			SubnetGUID:       subnetGUID,
			VnetGUID:         vnetGUID,
			SubnetResourceID: subnetResourceID,
		},
	}

	salOverrideToken = strings.TrimSpace(salOverrideToken)
	if len(salOverrideToken) != 0 {
		podNetwork.ObjectMeta.Labels = map[string]string{
			quarkconfig.AzureSALTokenLabelKey: salOverrideToken,
		}
	}

	return
}

// AzurePodNetworkCreate attempts to create a PodNetwork CR
func (client *NVCClient) AzurePodNetworkCreate(
	podNetwork *acnv1alpha1.PodNetwork,
) (*acnv1alpha1.PodNetwork, error) {
	if err := client.ctrlKubeClient.Create(context.TODO(), podNetwork); err != nil {
		return nil, err
	}

	return podNetwork, nil
}

// AzurePodNetworkDelete attempts to delete a PodNetwork CR
func (client *NVCClient) AzurePodNetworkDelete(
	podNetworkName string,
) error {
	podNetwork := &acnv1alpha1.PodNetwork{}
	podNetwork.Name = podNetworkName

	// Delete PN object and return
	return client.ctrlKubeClient.Delete(context.TODO(), podNetwork)
}

// AzurePodNetworkGet attempts to get a PodNetwork CR
func (client *NVCClient) AzurePodNetworkGet(
	podNetworkName string,
) (*acnv1alpha1.PodNetwork, error) {
	podNetwork := &acnv1alpha1.PodNetwork{}
	if err := client.ctrlKubeClient.Get(
		context.TODO(),
		ctrlClient.ObjectKey{Name: podNetworkName},
		podNetwork,
	); err != nil {
		return nil, err
	}

	return podNetwork, nil
}

// AzurePodNetworkList attempts to list all PodNetwork CRs
func (client *NVCClient) AzurePodNetworkList() (*acnv1alpha1.PodNetworkList, error) {
	podNetworkList := &acnv1alpha1.PodNetworkList{}
	if err := client.ctrlKubeClient.List(context.TODO(), podNetworkList); err != nil {
		return nil, err
	}

	return podNetworkList, nil
}
