package v2

import (
	"context"
	"encoding/json"
	"fmt"
	"strings"

	"github.com/google/go-cmp/cmp"
	"helm.sh/helm/v3/pkg/release"
	corev1 "k8s.io/api/core/v1"
	rbachelpers "k8s.io/api/rbac/v1"

	yaml "gopkg.in/yaml.v3"

	// API machinery
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	types "k8s.io/apimachinery/pkg/types"

	// APIExtensions
	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"

	// quark APIs
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
	"quark.netapp.io/volume-controller/pkg/util/helm"
)

const (
	nodevolLabel = "app=nodevol"
)

type patchStringValue struct {
	Op    string `json:"op"`
	Path  string `json:"path"`
	Value string `json:"value"`
}

func (client *NVCClient) ReconcileQuarkOperatorViaHelm(
	base64EncodedChart string, values map[string]interface{},
) error {
	err := helm.NewHelmReconciler(client.kubeConfig,
		base64EncodedChart,
		values,
		helm.QuarkOpChartReleaseName,
		client.nvcNamespace).Reconcile()
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return StaleVersionError{}
		} else {
			return err
		}
	}
	return nil
}

// GetReleaseViaHelm returns an installed helm release for the given chart name if it exists
// if it does not exist nil will be returned with no error, if an error occurs it will be returned
func (client *NVCClient) GetReleaseViaHelm(chartName helm.QuarkHelmChart) (chart *release.Release, err error) {
	return helm.NewHelmReconciler(client.kubeConfig,
		"",
		nil,
		chartName,
		client.nvcNamespace).GetInstalledRelease()
}

func (client *NVCClient) reconcileServiceAccount(sa *corev1.ServiceAccount) error {
	saCurr, err := client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Get(
		context.TODO(), sa.Name, metav1.GetOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			_, err := client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Create(
				context.TODO(), sa, metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create service account: %s, err: %s", sa.Name, err.Error())
			}
		} else {
			return fmt.Errorf("failed to get service account: %s, err: %s", sa.Name, err.Error())
		}
	} else {
		// We are Only supporting a Single Image-Pull-Secret
		doPatch := false
		var oldSecretName, newSecretName string
		var payload []patchStringValue

		if len(saCurr.ImagePullSecrets) > 0 {
			oldSecretName = saCurr.ImagePullSecrets[0].Name
		}
		if len(sa.ImagePullSecrets) > 0 {
			newSecretName = sa.ImagePullSecrets[0].Name
		}
		if oldSecretName != "" && newSecretName != "" && oldSecretName != newSecretName {
			doPatch = true
			// we are here because we originally had a different secret
			// now we are replacing that
			payload = []patchStringValue{{
				Op:    "replace",
				Path:  "/imagePullSecrets/0/name",
				Value: newSecretName,
			}}
		}

		if doPatch {
			payloadBytes, err := json.Marshal(payload)
			if err != nil {
				return fmt.Errorf("failed to marshal payload for patching image-pull-secret")
			}
			_, err = client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Patch(
				context.TODO(), sa.Name, types.JSONPatchType, payloadBytes, metav1.PatchOptions{})
			if err != nil {
				return fmt.Errorf("failed to patch image-prepull-secret for service account: %s, err: %s", sa.Name, err.Error())
			}
		} else if oldSecretName != newSecretName {
			_, err = client.kubeClient.CoreV1().ServiceAccounts(sa.Namespace).Update(
				context.TODO(), sa, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update image-prepull-secret for service account: %s, err: %s", sa.Name, err.Error())
			}
		}
	}
	return nil
}

// Creates and returns quark-operator ClusterRole
func createQuarkOperatorClusterRole() *rbachelpers.ClusterRole {
	return &rbachelpers.ClusterRole{
		ObjectMeta: metav1.ObjectMeta{
			Name: quarkconfig.QuarkOperatorSA,
		},
		Rules: []rbachelpers.PolicyRule{
			{
				APIGroups: []string{""},
				Resources: []string{"serviceaccounts"},
				Verbs:     []string{"get", "list", "create", "delete", "watch", "patch", "update"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"pods"},
				Verbs:     []string{"get", "list", "watch", "update", "patch", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"nodes"},
				Verbs:     []string{"get", "list", "watch", "update", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"nodes/proxy", "nodes/metrics"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"nodes/status"},
				Verbs:     []string{"get", "list", "watch", "update"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"ingresses"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"persistentvolumes"},
				Verbs:     []string{"get", "create", "list", "watch", "update", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"persistentvolumeclaims"},
				Verbs:     []string{"get", "create", "list", "watch", "update", "delete", "patch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"configmaps", "endpoints", "events"},
				Verbs:     []string{"get", "list", "create", "watch", "update", "delete", "patch"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"services"},
				Verbs:     []string{"get", "list", "create", "watch", "update", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"namespaces"},
				Verbs:     []string{"get", "list", "create", "watch", "delete"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"limitranges", "replicationcontrollers", "resourcequotas"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"apps"},
				Resources: []string{"deployments", "daemonsets", "replicasets", "statefulsets"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"apps"},
				Resources: []string{"deployments"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"etcd.database.coreos.com"},
				Resources: []string{"etcdclusters", "etcdbackups", "etcdrestores"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{""},
				Resources: []string{"pods", "endpoints", "secrets", "persistentvolumeclaims", "events"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"storage.k8s.io"},
				Resources: []string{"storageclasses"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete", "patch"},
			},
			{
				APIGroups: []string{"coordination.k8s.io"},
				Resources: []string{"leases"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"snapshot.storage.k8s.io"},
				Resources: []string{"volumesnapshots"},
				Verbs:     []string{"get", "list"},
			},
			{
				APIGroups: []string{"apiextensions.k8s.io"},
				Resources: []string{"customresourcedefinitions", "customresourcedefinitions/status"},
				Verbs:     []string{"*"},
			},
			{
				APIGroups: []string{"rbac.authorization.k8s.io"},
				Resources: []string{"roles", "rolebindings", "clusterroles", "clusterrolebindings"},
				Verbs:     []string{"get", "create", "delete", "update"},
			},
			{
				APIGroups: []string{"admissionregistration.k8s.io"},
				Resources: []string{"mutatingwebhookconfigurations", "validatingwebhookconfigurations"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"authentication.k8s.io"},
				Resources: []string{"tokenreviews"},
				Verbs:     []string{"create"},
			},
			{
				APIGroups: []string{"authorization.k8s.io"},
				Resources: []string{"subjectaccessreviews"},
				Verbs:     []string{"create"},
			},
			{
				APIGroups: []string{"autoscaling"},
				Resources: []string{"horizontalpodautoscalers"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"storage.k8s.io"},
				Resources: []string{"volumeattachments"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"batch"},
				Resources: []string{"cronjobs", "jobs"},
				Verbs:     []string{"get", "list", "watch", "create", "update", "delete"},
			},
			{
				APIGroups: []string{"certificates.k8s.io"},
				Resources: []string{"certificatesigningrequests"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"networking.k8s.io"},
				Resources: []string{"ingress", "ingresses", "ingresses/status"},
				Verbs:     []string{"get", "list", "watch"},
			},
			{
				APIGroups: []string{"networking.k8s.io"},
				Resources: []string{"networkpolicies"},
				Verbs:     []string{"list", "watch"},
			},
			{
				APIGroups: []string{"policy"},
				Resources: []string{"poddisruptionbudgets"},
				Verbs: []string{
					"list", "watch", "get", "create", "delete",
				},
			},
			{
				APIGroups: []string{"scheduling.k8s.io"},
				Resources: []string{"priorityclasses"},
				Verbs: []string{
					"get", "create", "delete",
				},
			},
			{
				NonResourceURLs: []string{"/metrics"},
				Verbs:           []string{"get"},
			},
			{
				APIGroups: []string{quarkconfig.QuarkAPIGroupName},
				Resources: []string{"*"},
				Verbs:     []string{"*"},
			},
		},
	}
}

// ReconcileQuarkOperatorRBAC publicly exposes Quark Operator RBAC reconciliation for use by SDE in updating their default service accounts
func (client *NVCClient) ReconcileQuarkOperatorRBAC(pullSecret, registry, namespace string) error {
	return client.reconcileQuarkOperatorRBAC(pullSecret, registry, namespace)
}

func (client *NVCClient) reconcileQuarkOperatorRBAC(pullSecret, registry, namespace string) error {
	// Creation of Quark operator SA, Cluster Role and Cluster Role Bindings
	// is no longer needed here as the logic has been moved to Helm.
	// TODO: Remove the following part where we patch the default service
	// account with imagePullSecret. Currently Etcd cluster pods use the
	// default service account.

	// Patch default service account in QuarkSystemNamespace with ImagePullSecrets
	sa := &corev1.ServiceAccount{
		ObjectMeta: metav1.ObjectMeta{
			Name:      "default",
			Namespace: quarkconfig.QuarkSystemNamespace,
		},
	}
	if pullSecret != "" {
		sa.ImagePullSecrets = []corev1.LocalObjectReference{{Name: pullSecret}}
	}

	err := client.reconcileServiceAccount(sa)
	if err != nil {
		return fmt.Errorf("failed to reconcile service account %s: %s", sa.Name, err.Error())
	}
	return nil
}

func makeQuarkCRD() *apiextensionsv1.CustomResourceDefinition {
	preserveUnknownFields := true
	qCrd := &apiextensionsv1.CustomResourceDefinition{
		ObjectMeta: metav1.ObjectMeta{
			Name: quarkconfig.QuarkCRDName,
		},
		Spec: apiextensionsv1.CustomResourceDefinitionSpec{
			Group: quarkconfig.QuarkAPIGroupName,
			Scope: apiextensionsv1.NamespaceScoped,

			Versions: []apiextensionsv1.CustomResourceDefinitionVersion{
				{
					Name:    "v1alpha1",
					Served:  true,
					Storage: true,
					Schema: &apiextensionsv1.CustomResourceValidation{
						OpenAPIV3Schema: &apiextensionsv1.JSONSchemaProps{
							Type: "object",
							Properties: map[string]apiextensionsv1.JSONSchemaProps{
								"apiVersion": {
									Type: "string",
								},
								"kind": {
									Type: "string",
								},
								"spec": {
									Type:                   "object",
									XPreserveUnknownFields: &preserveUnknownFields,
								},
								"status": {
									Type:                   "object",
									XPreserveUnknownFields: &preserveUnknownFields,
								},
							},
						},
					},
					Subresources: &apiextensionsv1.CustomResourceSubresources{
						Status: &apiextensionsv1.CustomResourceSubresourceStatus{},
					},
					AdditionalPrinterColumns: []apiextensionsv1.CustomResourceColumnDefinition{
						{
							Name:        "Version",
							Type:        "string",
							Description: "Quark Spec Version",
							JSONPath:    ".spec.version",
						},
						{
							Name:        "Status",
							Type:        "string",
							Description: "Quark Status Version",
							JSONPath:    ".status.version",
						},
					},
				},
			},
			Names: apiextensionsv1.CustomResourceDefinitionNames{
				Kind:     "Quark",
				ListKind: "QuarkList",
				Plural:   "quarks",
				Singular: "quark",
			},
		},
	}

	return qCrd
}

// NetAppQuarkOperatorReadyforEvent will return if Quark Operator is in ready state for Events
func (client *NVCClient) NetAppQuarkOperatorReadyforEvent(quarkCR *quarkv1alpha1.Quark) bool {
	quarkStats := getNetAppQuarkStatus(quarkCR)
	// check for quark-operator
	return quarkStats.OperatorReady
}

// NetAppQuarkOperatorReady will return if Quark Operator is in ready state with new image
func (client *NVCClient) NetAppQuarkOperatorReady(quarkCR *quarkv1alpha1.Quark, newVersion string) (bool, error) {
	// verify whether quark operator has been reconciled before updating quark version
	quarkopVersion, err := client.getQuarkOperatorVersion()
	if err != nil {
		return false, fmt.Errorf("failed to get operator version: %s", err.Error())
	}
	if quarkopVersion != newVersion {
		return false, fmt.Errorf("reconcile quark operator with the new version and try again")
	}

	// Verify if the quark-operator pod is running with new image
	// additional check for cases where deployment controller is updated but pod has not been recreated #burt1488474
	if err = client.isQuarkOperatorPodUpdated(newVersion); err != nil {
		return false, fmt.Errorf("failed to validate quark operator pod is updated: %s", err.Error())
	}

	return true, nil
}

// getHelmTag looks up a json path value(e.g. root.field1.field1a) on a helm chart and returns it, if it is present and a string type
// otherwise, it returns empty string
func getHelmTag(mapValues map[string]interface{}, path string) string {
	chartValue, chartValueFound, err := helm.GetPathValue(path, mapValues)
	if err != nil || !chartValueFound {
		return ""
	}
	chartValueStr, ok := chartValue.(string)
	if !ok {
		return ""
	}
	return chartValueStr
}

// NetAppQuarkOperatorReadyViaHelm checks if the quark operator is running and reconciled to the correct version after install / upgrade
// instead of an image tag it takes the same paramters passed to the operator reconcile function to determine the correct image tag/digest to look for
func (client *NVCClient) NetAppQuarkOperatorReadyViaHelm(helmChart string, values map[string]interface{}) (bool, error) {
	helmDefaultValues, err := helm.GetValuesMapFromChart("values.yaml", helmChart)
	if err != nil {
		return false, err
	}
	// Attempt to get (in order) correct image from the override values, or defaults from the helm chart
	// Priority goes to a digest first, then tag
	helmDefaultTag := getHelmTag(helmDefaultValues, "images.quarkop.tag")
	helmDefaultDigest := getHelmTag(helmDefaultValues, "images.quarkop.digest")
	overrideTag := getHelmTag(values, "images.quarkop.tag")
	overrideDigest := getHelmTag(values, "images.quarkop.digest")

	tags := []string{overrideDigest, overrideTag, helmDefaultDigest, helmDefaultTag}
	targetImageTag := ""
	for _, tag := range tags {
		if tag != "" {
			targetImageTag = tag
			break
		}
	}
	if targetImageTag == "" {
		return false, fmt.Errorf("unable to get operator image from helm chart defaults or from override values")
	}
	return client.NetAppQuarkOperatorReady(nil, targetImageTag)
}

// NetAppQuarkReady will return True if all the Pods are running with latest image version
// defined in the Quark.Spec.Images and all the Quark Components are Healthy
func (client *NVCClient) NetAppQuarkReady(quarkCR *quarkv1alpha1.Quark) bool {
	// Check for quark version
	if quarkCR.Spec.Version != quarkCR.Status.Version {
		fmt.Println("Quark has not reconciled yet")
		return false
	}

	// Replica count
	replicaCount := 1
	nhcReplicasCount := 2
	if quarkCR.Spec.Regional {
		nhcReplicasCount = 3
	}
	quarkDS, err := client.GetKubeClient().AppsV1().DaemonSets(quarkconfig.QuarkSystemNamespace).Get(
		context.TODO(), "image-prepull", metav1.GetOptions{})
	if err != nil {
		// Should this be logger?
		fmt.Printf("error fetching daemonset %v", err.Error())
		return false
	}
	dsReplicasCount := int(quarkDS.Status.DesiredNumberScheduled)
	etcdReplicasCount := quarkCR.Spec.ControllerConfig.EtcdClusterSize

	// Pods & Containers to Check
	podsInfo := []quarkconfig.PodInfo{
		{
			Namespace: client.nvcNamespace, Replica: replicaCount, PodLabelSelector: "app=netapp-volume-controller",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "controller", Image: quarkCR.Spec.Images.NVC},
			},
		},
		{
			Namespace: quarkconfig.QuarkSystemNamespace, Replica: etcdReplicasCount, PodLabelSelector: "app=etcd",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "etcd", Image: quarkCR.Spec.Images.Etcd},
			},
		},
		{
			Namespace: quarkconfig.QuarkSystemNamespace, Replica: replicaCount, PodLabelSelector: "app=nodevol-controller",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "controller", Image: quarkCR.Spec.Images.NodevolController},
			},
		},
		{
			Namespace: quarkconfig.QuarkSystemNamespace, Replica: dsReplicasCount, PodLabelSelector: nodevolLabel,
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "supportability", Image: quarkCR.Spec.Images.Supportability},
			},
		},
		{
			Namespace: quarkconfig.QuarkSystemNamespace, Replica: dsReplicasCount, PodLabelSelector: "prepull=primary",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "etcd", Image: quarkCR.Spec.Images.Etcd},
			},
		},
	}

	// We do not deploy HA Controller and node monitor for Min config.
	if quarkCR.Spec.ControllerConfig.Variant != quarkconfig.MinVariant {
		podsInfo = append(podsInfo, quarkconfig.PodInfo{
			Namespace: client.nvcNamespace, Replica: nhcReplicasCount, PodLabelSelector: "app=netapp-ha-controller",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "controller", Image: quarkCR.Spec.Images.NHC},
			},
		},
			quarkconfig.PodInfo{
				Namespace: quarkconfig.QuarkSystemNamespace, Replica: dsReplicasCount, PodLabelSelector: "app=node-monitor",
				Container: map[string]quarkconfig.ContainerImage{
					"container": {Name: "node-monitor", Image: quarkCR.Spec.Images.NodeMonitor},
				},
			})
	}

	// We do not deploy ntapvol-snat-controller for azure when swift is enabled.
	if !(quarkCR.Spec.Cloud == quarkconfig.CloudAzure && quarkCR.Spec.EnableSWIFTNetworking) {
		podsInfo = append(podsInfo, quarkconfig.PodInfo{
			Namespace: quarkconfig.QuarkSystemNamespace, Replica: dsReplicasCount, PodLabelSelector: "app=ntapvol-daemonset",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "ntapvol-snat-controller", Image: quarkCR.Spec.Images.NSC},
			},
		})
	}

	// Add cloud snapshot check if the user had set csc or adc images.
	// These both should be either empty or populated, but this checks for either being set to prevent quark ready from hiding bugs
	if quarkCR.Spec.Images.CSC != "" || quarkCR.Spec.Images.ADC != "" {
		podsInfo = append(podsInfo, quarkconfig.PodInfo{
			Namespace: client.nvcNamespace, Replica: replicaCount, PodLabelSelector: "app=cloud-snapshot-controller",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "controller", Image: quarkCR.Spec.Images.CSC},
			},
		},
		)
	}
	// service-mesh-controller pod info
	if quarkCR.Spec.EnableCrr && quarkCR.Spec.Cloud != quarkconfig.CloudAzure {
		smcPodInfo := quarkconfig.PodInfo{
			Namespace: client.nvcNamespace, Replica: replicaCount, PodLabelSelector: "app=asm-controller",
			Container: map[string]quarkconfig.ContainerImage{
				"container": {Name: "svcmesh-controller", Image: quarkCR.Spec.Images.SvcMeshCtrl},
			},
		}
		podsInfo = append(podsInfo, smcPodInfo)
	}
	for _, podInfo := range podsInfo {
		if !client.isPodUpdated(podInfo.PodLabelSelector, podInfo.Namespace, podInfo.Container["container"].Name,
			podInfo.Container["container"].Image, podInfo.Replica) {
			return false
		}
	}

	// Get the Health Status after all pods running with new image
	quarkStats := getNetAppQuarkStatus(quarkCR)
	// check netapp-volume-controller health
	if !quarkStats.VolumeControllerReady {
		fmt.Println("netapp-volume-controller not ready yet")
		return false
	}

	// check for image-prepull health
	if !quarkStats.IPDReady {
		fmt.Println("image-prepull not ready yet")
		return false
	}
	if quarkCR.Spec.ControllerConfig.Variant != quarkconfig.MinVariant {
		// check netapp-ha-controllers health
		if quarkCR.Spec.Cloud == quarkconfig.CloudGCP && !quarkStats.HighAvailabilityReady {
			fmt.Println("netapp-ha-controllers not ready yet")
			return false
		}
		// check for node-monitor health
		if !quarkStats.NodeMonitorReady {
			fmt.Println("node-monitor not ready yet")
			return false
		}
	}
	// check for etcd-cluster health
	if !quarkStats.EtcdClusterReady {
		fmt.Println("etcd-cluster not ready yet")
		return false
	}

	// check for node-vol health
	if strings.EqualFold(quarkCR.Spec.Nodevol, "true") {
		if !quarkStats.NodevolReady {
			fmt.Println("node-vol not ready yet")
			return false
		}
	}
	// check for ntapvol-snat-controller health
	if !quarkStats.ExternalVolumeAccessReady {
		fmt.Println("ntapvol-snat-controller not ready yet")
		return false
	}
	// check for cloud-snapshot-controller health
	if !quarkStats.CloudsnapshotCapabilityReady {
		fmt.Println("cloud-snapshot-controller not ready yet")
		return false
	}
	// check for service-mesh-controller health
	if quarkCR.Spec.Cloud == quarkconfig.CloudGCP && quarkCR.Spec.EnableCrr {
		if !quarkStats.ServiceMeshControllerReady {
			fmt.Println("service-mesh-controller not ready yet")
			return false
		}
	}
	// check for otel-collector-ready.
	if quarkCR.Spec.EnableGcp1p || quarkCR.Spec.Cloud == quarkconfig.CloudAzure {
		if !quarkStats.OtelCollectorReady {
			fmt.Println("otel-collector not ready yet")
			return false
		}
	}
	return true
}

// NetAppQuarkThirdPartyReady will return true if Quark third party components are in ready state
func (client *NVCClient) NetAppQuarkThirdPartyComponentsReady(quarkCR *quarkv1alpha1.Quark) bool {
	quarkThirdPartyStats := getNetAppQuarkThirdPartyComponentStatus(quarkCR)

	compStatus := quarkThirdPartyStats.PrometheusServerReady && quarkThirdPartyStats.KubeStateMetricsReady && quarkThirdPartyStats.ADCReady

	return compStatus
}

// NetAppQuarkGetFirst will get the first Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkGetFirst() (*quarkv1alpha1.Quark, error) {
	quarks, err := client.NetAppQuarkGetAll()
	if len(quarks) == 0 {
		return nil, fmt.Errorf("failed to get quark in namespace '%s'", client.nvcNamespace)
	}

	return quarks[0], err
}

// NetAppQuarkGet will get Quark CR in all the Namespaces in the cluster
func (client *NVCClient) NetAppQuarkGet(quarkCRName string) (*quarkv1alpha1.Quark, error) {
	quarkList, err := client.NetAppQuarkGetAll()
	if err != nil {
		return nil, err
	}
	return quarkList[0], nil
}

// NetAppQuarkGetAll will get all Quark CRs in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkGetAll() ([]*quarkv1alpha1.Quark, error) {
	quarks := make([]*quarkv1alpha1.Quark, 0)
	result, err := client.listWithRetry(&quarkv1alpha1.QuarkList{}, metav1.ListOptions{})
	if err != nil {
		if errors.IsNotFound(err) {
			return nil, err
		}
		return nil, fmt.Errorf("failed to list quarks in cluster: %s", err.Error())
	}
	quarkList := result.(*quarkv1alpha1.QuarkList)
	for i := range quarkList.Items {
		quarks = append(quarks, &quarkList.Items[i])
	}
	return quarks, nil
}

// NetAppQuarkCreate will attempt to create a Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkCreate(quarkCR *quarkv1alpha1.Quark) (*quarkv1alpha1.Quark, error) {
	result, err := client.listWithRetry(&quarkv1alpha1.QuarkList{}, metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to list quark in cluster: %s", err.Error())
	}
	quarkList := result.(*quarkv1alpha1.QuarkList)

	// Get List of Quarks in cluster and fail new creation if one exists
	if len(quarkList.Items) != 0 {
		if strings.EqualFold(quarkList.Items[0].Name, quarkCR.Name) {
			return nil, fmt.Errorf("failed to create quark, %s already exists", quarkCR.Name)
		}
		return nil, fmt.Errorf("failed to create quark, operation not allowed")
	}

	// Validations for QuarkCR
	switch quarkCR.Spec.Cloud {
	case quarkconfig.CloudGCP:
	case quarkconfig.CloudCTL:
	case quarkconfig.CloudAzure:
	default:
		return nil, fmt.Errorf("failed to create quark, invalid cloud specified")
	}

	// Validations for Variant
	switch quarkCR.Spec.ControllerConfig.Variant {
	case quarkconfig.ProductionVariant:
	case quarkconfig.OnPremVariant:
	default:
		return nil, fmt.Errorf("failed to create quark, invalid variant")
	}

	result, err = client.createWithRetry(quarkCR)
	if err != nil {
		return nil, err
	}
	return result.(*quarkv1alpha1.Quark), nil
}

func (client *NVCClient) ReconcileQuarkCRViaHelm(
	base64EncodedChart string, values map[string]interface{},
) (*quarkv1alpha1.Quark, error) {
	err := helm.NewHelmReconciler(client.kubeConfig,
		base64EncodedChart,
		values,
		helm.QuarkCRReleaseName,
		client.nvcNamespace).Reconcile()
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		} else {
			return nil, err
		}
	}
	result, err := client.listWithRetry(&quarkv1alpha1.QuarkList{}, metav1.ListOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to list quark in cluster: %s", err.Error())
	}
	quarks := result.(*quarkv1alpha1.QuarkList)
	if len(quarks.Items) != 1 {
		return nil, fmt.Errorf("%d number of quark objects found after ReconcileQuarkCRHelm, there should only be one", len(quarks.Items))
	}
	return &quarks.Items[0], nil
}

// Function to set feature flags true for upgrade
func (client *NVCClient) setFeatureFlags(featureFlags []string) quarkv1alpha1.QuarkConfigParams {
	var featureFlagsTrimmed = make([]string, len(featureFlags))
	for i := range featureFlags {
		featureFlagsTrimmed[i] = strings.TrimSpace(featureFlags[i])
	}

	enableFlags := quarkv1alpha1.QuarkConfigParams{
		SetNfs41:               util.StringSetContains(featureFlagsTrimmed, "enableNfs41"),
		SetBackupSecurityUsers: util.StringSetContains(featureFlagsTrimmed, "enableBackupSecurityUsers"),
		SetEtcdCostReduction:   util.StringSetContains(featureFlagsTrimmed, "enableEtcdCostReduction"),
		SetCcpd:                util.StringSetContains(featureFlagsTrimmed, "enableCcpd"),
		SetCrr:                 util.StringSetContains(featureFlagsTrimmed, "enableCrr"),
		SetServiceMesh:         util.StringSetContains(featureFlagsTrimmed, "enableServiceMesh"),
		SetGranularLiveness:    util.StringSetContains(featureFlagsTrimmed, "enableGranularLiveness"),
		SetIntegratedBackup:    util.StringSetContains(featureFlagsTrimmed, "enableIntegratedBackup"),
	}

	return enableFlags
}

func (client *NVCClient) getQuarkOperatorVersion() (string, error) {
	quarkOpDeploy, err := client.kubeClient.AppsV1().Deployments(client.nvcNamespace).Get(
		context.TODO(), quarkconfig.QuarkOperatorDeploymentName, metav1.GetOptions{})
	if err != nil {
		return "", err
	}
	containers := quarkOpDeploy.Spec.Template.Spec.Containers
	// quarkop deployment has only one container and container image is of the format "<REGISTRY>/quark-operator:<VERSION>"
	return util.GetImageTagOrDigest(containers[0].Image)

}

func (client *NVCClient) isQuarkOperatorPodUpdated(newVersion string) error {
	options := metav1.ListOptions{
		LabelSelector: "app=quark-operator",
	}
	quarkOpPodList, err := client.GetKubeClient().CoreV1().Pods(client.nvcNamespace).List(context.TODO(), options)
	if err != nil {
		return err
	}
	if len(quarkOpPodList.Items) == 0 || !util.IsPodRunning(&quarkOpPodList.Items[0]) {
		return fmt.Errorf("quark operator pod is not running")
	}
	if len(quarkOpPodList.Items[0].Spec.Containers) == 0 {
		return fmt.Errorf("quark operator pod's container spec is missing")
	}
	// Check spec of operator container
	opContainerTag, err := util.GetImageTagOrDigest(quarkOpPodList.Items[0].Spec.Containers[0].Image)
	if err != nil {
		return err
	}
	if opContainerTag != newVersion {
		return fmt.Errorf("quark operator pod's container spec is not yet updated with latest image")
	}
	if len(quarkOpPodList.Items[0].Status.ContainerStatuses) == 0 {
		return fmt.Errorf("the quark operator pod's container status is missing")
	}
	// Check status of operator container
	if !quarkOpPodList.Items[0].Status.ContainerStatuses[0].Ready {
		return fmt.Errorf("quark operator pod's container status is not yet ready. Current State is %s",
			quarkOpPodList.Items[0].Status.ContainerStatuses[0].State)
	}
	return nil
}

func (client *NVCClient) isPodUpdated(label string, namespace string, containerName string, newImage string, replicaCount int) bool {
	options := metav1.ListOptions{
		LabelSelector: label,
	}
	// Get the Pods List
	podList, err := client.GetKubeClient().CoreV1().Pods(namespace).List(context.TODO(), options)
	if err != nil {
		fmt.Printf("Error listing pods using selector %v: %v\n", label, err)
		return false
	}
	//check length of podList for pods
	podCount := len(podList.Items)
	if label == nodevolLabel {
		if podCount < replicaCount {
			fmt.Printf("Found %v replica pods for selector %v, but expected more than or equal to %v\n", podCount, label, replicaCount)
			return false
		}
	} else if podCount != replicaCount {
		fmt.Printf("Found %v replica pods for selector %v, but expected %v\n", podCount, label, replicaCount)
		return false
	}
	// Iterate over Pods List
	for _, podListItem := range podList.Items {
		// Validate Pod is running
		if !util.IsPodRunning(&podListItem) {
			fmt.Printf("Pod %v for label selector %v is not running\n", podListItem.Name, label)
			return false
		}
		// Check the length of containers spec
		if len(podListItem.Spec.Containers) == 0 {
			fmt.Printf("Spec of Pod %v has not been updated with Container info\n", podListItem.Name)
			return false
		}
		// Check spec image of controller container
		for _, container := range podListItem.Spec.Containers {
			if (container.Name == containerName) && (container.Image != newImage) {
				fmt.Printf("Container %v for pod %v has image %v while desired image is %v\n",
					container.Name, podListItem.Name, container.Image, newImage)
				return false
			}
		}
		// Check the length of containers status
		if len(podListItem.Status.ContainerStatuses) == 0 {
			fmt.Printf("Pod %s container status not updated\n", label)
			return false
		}
		// Check status of controller container
		for _, containerStatus := range podListItem.Status.ContainerStatuses {
			if (containerStatus.Name == containerName) && !containerStatus.Ready {
				fmt.Printf("Container %v for pod %v has container status image %v when %v is desired: ready status is %v\n",
					containerStatus.Name, podListItem.Name, containerStatus.Image, newImage, containerStatus.Ready)
				return false
			}
		}
	}
	return true
}

// Patch Specific quark or third party images in quark
func (client *NVCClient) NetAppQuarkImagePatch(quarkCRName string, imageYaml string, registry string) (*quarkv1alpha1.Quark, error) {
	quarkCR, err := client.NetAppQuarkGet(quarkCRName)
	if err != nil {
		return nil, fmt.Errorf("failed to get quark: %s", err.Error())
	}

	if registry == "" {
		// Get the existing registry from the dmap image
		registry, err = util.GetRegistryFromImagePath(quarkCR.Spec.Images.DMAP)
		if err != nil {
			return nil, err
		}
	}

	// Unmarshal yaml
	images, err := client.GetImageMapFromImageYaml(imageYaml, registry)
	if err != nil {
		return nil, err
	}

	// Patch all images based on the image yaml that is passed in
	quarkCRCopy, err := patchImages(quarkCR, images, registry)
	if err != nil {
		return nil, err
	}

	newQuark, err := client.updateWithRetry(quarkCRCopy)
	if err != nil {
		if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
			return nil, StaleVersionError{}
		}
		return nil, err
	}

	return newQuark.(*quarkv1alpha1.Quark), nil
}

// NetAppQuarkDelete will attempt to delete the given Quark CR in the nvcClient Namespace in the cluster
func (client *NVCClient) NetAppQuarkDelete(quarkCRName string) error {
	vols, err := client.NetAppVolumeGetAll()
	if err != nil {
		return fmt.Errorf("failed to get netappvolumes: %s", err.Error())
	}

	if len(vols) != 0 {
		return fmt.Errorf("%v netappvolumes exist in cluster, cannot delete quark", len(vols))
	}
	quark := &quarkv1alpha1.Quark{
		ObjectMeta: metav1.ObjectMeta{
			Name: quarkCRName,
		},
	}
	return client.deleteWithRetry(quark)
}

func (client *NVCClient) GetImageMapFromImageYaml(imageYaml string, registry string) (map[string]interface{}, error) {
	images := make(map[string]interface{})
	err := yaml.Unmarshal([]byte(imageYaml), &images)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal images yaml %v", err)
	}

	// append registry to images
	for componentName, imagetag := range images {
		images[componentName] = registry + "/" + imagetag.(string)
	}

	return images, nil
}

// Api that returns an image struct given an image YAML
func (client *NVCClient) GetImageStructFromImageYaml(imageYaml string, registry string) (quarkv1alpha1.Images, error) {
	var images quarkv1alpha1.Images

	imageMap, err := client.GetImageMapFromImageYaml(imageYaml, registry)
	if err != nil {
		return images, err
	}
	err = util.FillStructFieldsFromMap(&images, imageMap)
	return images, err
}

func (client *NVCClient) NetAppQuarkPostUpgradeCleanup(quarkCR *quarkv1alpha1.Quark) error {
	// check if quark version for spec and status match
	if quarkCR.Spec.Version != quarkCR.Status.Version {
		return fmt.Errorf("quark is not upgraded: %s", quarkCR.Spec.Version)
	}

	// check if all SPs are upgraded
	storagePools, err := client.NetAppStoragePoolGetAll()
	if err != nil {
		return fmt.Errorf("unable to get existing storage pools: %v", err)
	}

	if len(storagePools) > 0 {
		var nonUpgradedSP []string
		for _, sp := range storagePools {
			if !client.NetAppStoragePoolUpgraded(sp) || sp.Status.QuarkVersion != quarkCR.Status.Version {
				nonUpgradedSP = append(nonUpgradedSP, sp.Name)
			}
		}
		if len(nonUpgradedSP) > 0 {
			return fmt.Errorf("storage pools not upgraded to the current version of quark: %s", strings.Join(nonUpgradedSP, ", "))
		}
	}

	// post upgrade cleanup
	// delete backup ipd ds
	if err = client.kubeClient.AppsV1().DaemonSets(quarkconfig.QuarkSystemNamespace).Delete(
		context.TODO(),
		quarkconfig.BackupPrepullDaemonSet,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete daemonset: %s, %v", quarkconfig.BackupPrepullDaemonSet, err)
	}

	// delete old secrets
	oldSecrets := map[string][]string{
		client.nvcNamespace:              quarkconfig.PurgeOldSecretsNSDefault,
		quarkconfig.QuarkSystemNamespace: quarkconfig.PurgeOldSecretsNSQuarkSystem,
	}

	for ns, secrets := range oldSecrets {
		for _, secretName := range secrets {
			if _, err := client.kubeClient.CoreV1().Secrets(ns).Get(
				context.TODO(), secretName, metav1.GetOptions{}); err != nil && !errors.IsNotFound(err) {
				return fmt.Errorf("failed to get secret (%s) and error : (%v)", secretName, err)
			}

			if err := client.kubeClient.CoreV1().Secrets(ns).Delete(
				context.TODO(), secretName, metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
				return fmt.Errorf("failed to delete secret (%s) and error (%v)", secretName, err)
			}
		}
	}

	// Delete old rbac roles and bindings
	if err := client.kubeClient.RbacV1().ClusterRoleBindings().Delete(
		context.TODO(),
		quarkconfig.NetAppPodrickClusterRoleBindingOld,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete cluster role binding: %s, %v", quarkconfig.NetAppPodrickClusterRoleBindingOld, err)
	}

	if err := client.kubeClient.RbacV1().ClusterRoles().Delete(
		context.TODO(),
		quarkconfig.NetAppPodrickClusterRoleOld,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete cluster role: %s, %v", quarkconfig.NetAppPodrickClusterRoleOld, err)
	}

	if err := client.kubeClient.RbacV1().RoleBindings(quarkCR.Namespace).Delete(
		context.TODO(),
		quarkconfig.NetAppPodrickRoleBindingOld,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete role binding: %s, %v", quarkconfig.NetAppPodrickRoleBindingOld, err)
	}

	if err := client.kubeClient.RbacV1().Roles(quarkCR.Namespace).Delete(
		context.TODO(),
		quarkconfig.NetAppPodrickRoleOld,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete role: %s, %v", quarkconfig.NetAppPodrickRoleOld, err)
	}

	if err := client.kubeClient.CoreV1().ServiceAccounts(quarkCR.Namespace).Delete(
		context.TODO(),
		quarkconfig.NetAppPodrickSAOld,
		metav1.DeleteOptions{}); err != nil && !errors.IsNotFound(err) {
		return fmt.Errorf("failed to delete service account: %s, %v", quarkconfig.NetAppPodrickSAOld, err)
	}

	return nil
}

// NetAppQuarkFeatureFlagReady compares feature flags in Quark CR Spec and Status and returns true if they are equal.
// It logs the difference between the Spec and Status feature flags if they are not equal.
func (client *NVCClient) NetAppQuarkFeatureFlagReady(quark *quarkv1alpha1.Quark) bool {
	if !util.CompareFeatureFlags(quark.Spec.FeatureFlags, quark.Status.FeatureFlags) {
		diff := cmp.Diff(quark.Spec.FeatureFlags, quark.Status.FeatureFlags)
		log.Warnf("Difference between Spec.FeatureFlags and Status.FeatureFlags:\n %s", diff)
		return false
	}
	return true
}

// NetAppStoragePoolFeatureFlagReady compares feature flags in StoragePool CR Spec and Status and returns true if they are equal.
// It logs the difference between the Spec and Status feature flags if they are not equal.
func (client *NVCClient) NetAppStoragePoolFeatureFlagReady(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if !util.CompareStoragePoolFeatureFlags(storagePool.Spec.FeatureFlags, storagePool.Status.FeatureFlags) {
		diff := cmp.Diff(storagePool.Spec.FeatureFlags, storagePool.Status.FeatureFlags)
		log.Warnf("Difference between Spec.FeatureFlags and Status.FeatureFlags:\n %s", diff)
		return false
	}
	return true
}

// IsStoragePoolFeatureFlagAlignedToQuark checks if the feature flags in the storage pool are aligned with the quark feature flags
// and returns true if they match. If they do not match, it logs the feature flags that are not aligned and returns false.
func (client *NVCClient) IsStoragePoolFeatureFlagAlignedToQuark(storagePool *quarkv1alpha1.NetAppStoragePool) (bool, error) {
	quark, err := client.NetAppQuarkGetFirst()
	if err != nil {
		return false, fmt.Errorf("NetAppQuarkGet failed: %v", err)
	}
	return areFeatureFlagsAligned(storagePool, quark), nil
}

func areFeatureFlagsAligned(storagePool *quarkv1alpha1.NetAppStoragePool, quark *quarkv1alpha1.Quark) bool {
	isAligned := true
	// Check for version mismatch
	if storagePool.Spec.QuarkVersion != quark.Spec.Version {
		log.Warnf("Quark version in storage pool is %s but it is %s in quark",
			storagePool.Spec.QuarkVersion, quark.Spec.Version)
	}
	spFeatureFlags := storagePool.Spec.FeatureFlags
	spFeatureFlagsInQuark := util.GetStoragePoolFeatureFlagsFromQuark(quark)
	// Get all the keys from both maps
	keys := make(map[quarkv1alpha1.FeatureFlagName]bool)
	for k := range spFeatureFlags {
		keys[k] = true
	}
	for k := range spFeatureFlagsInQuark {
		keys[k] = true
	}

	for key := range keys {
		value, spOk := spFeatureFlags[key]
		state, quarkOk := spFeatureFlagsInQuark[key]

		switch {
		case spOk && quarkOk:
			if value != state {
				log.Warnf("Feature flag %s in storage pool is %s but it is %s in quark",
					key, value, state)
				isAligned = false
			}
		case spOk:
			log.Warnf("Feature flag %s in storage pool is %s but it is not present in quark", key, value)
			isAligned = false
		case quarkOk:
			log.Warnf("Feature flag %s is not present in storage pool but it is %s in quark", key, state)
			isAligned = false
		}
	}

	return isAligned
}
