package v2

import (
	"fmt"
	"strings"

	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

func (client *NVCClient) ServiceAccountCreate(serviceAccountName string) error {
	// Used for internal development testing purpose only
	// SDE should create it's own function to create ServiceAccount

	// Create the ServiceAccount object
	serviceAccount := &corev1.ServiceAccount{
		ObjectMeta: metav1.ObjectMeta{
			Name: serviceAccountName,
		},
	}
	// Check if the ServiceAccount already exists
	_, err := client.getWithRetry(serviceAccount)
	if err != nil {
		if errors.IsNotFound(err) {
			// Create the ServiceAccount
			_, err := client.createWithRetry(serviceAccount)
			if err != nil {
				return err
			} else {
				log.Infof("ServiceAccount '%s' created\n", serviceAccountName)
			}
		} else {
			return fmt.Errorf("Unexpected error while trying to get ServiceAccount '%s': %v", serviceAccountName, err)
		}
	} else {
		log.Infof("ServiceAccount '%s' already exists\n", serviceAccountName)
	}
	return nil
}

func (client *NVCClient) ServiceAccountAnnotate(saInfo quarkconfig.CMEKServiceAccountDetails) error {
	// Used for internal development testing purpose only
	// It is being called internally while running Storage Pool Create and Encryption Update cmds
	// SDE should create it's own function to annotate ServiceAccount
	for {
		// Create the ServiceAccount object
		serviceAccount := &corev1.ServiceAccount{
			ObjectMeta: metav1.ObjectMeta{
				Name: saInfo.ServiceAccountName,
			},
		}
		// Check if the ServiceAccount already exists
		_, err := client.getWithRetry(serviceAccount)
		if err != nil {
			return err
		}
		// Check if the annotation is already present
		annotations := serviceAccount.Annotations
		if annotations == nil {
			annotations = make(map[string]string)
		}
		if value, ok := annotations[saInfo.AnnotationKey]; !ok || value != saInfo.AnnotationValue {
			// Add the annotation if it is not already present
			annotations[saInfo.AnnotationKey] = saInfo.AnnotationValue
			serviceAccount.Annotations = annotations

			// Update the ServiceAccount object
			_, err = client.updateWithRetry(serviceAccount)
			if err != nil {
				if strings.Contains(err.Error(), quarkconfig.OldVersionErr) {
					continue
				} else {
					return err
				}
			}
			log.Debugf("%s Annotation added to ServiceAccount '%s'\n", saInfo.AnnotationValue, saInfo.ServiceAccountName)
		} else {
			log.Debugf("%s Annotation already present on ServiceAccount '%s'\n", saInfo.AnnotationValue, saInfo.ServiceAccountName)
		}
		return nil
	}
}
