package v2

import (
	"context"
	"fmt"

	uuid "github.com/satori/go.uuid"
	"k8s.io/apimachinery/pkg/api/errors"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// nvrBackup implements the nvrInterface to support the backup usecase. Functionalities like break/resync which are not
// planned for the first release of Integrated backup feature should return error until implemented.
type nvrBackup struct {
	nvrCommonImpl
}

func (r *nvrBackup) Create() (*quarkv1alpha1.NetAppVolumeReplication, error) {
	storagePoolUUIDStr := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	_, err := r.nvcClient.NetAppStoragePoolGet(storagePoolUUIDStr)
	if err != nil {
		log.WithFields(log.Fields{
			"storagePool": storagePoolUUIDStr,
			"error":       err,
		}).Debug("error getting storage pool")
		return nil, err
	}

	objectStoreUUIDStr := r.nvr.Labels[quarkconfig.ObjectStoreLabel]
	// not validating err in the conversion below since if uuid string is part of labels in the cr, it means it has already been vetted earlier
	objectStoreUUID, _ := uuid.FromString(objectStoreUUIDStr)
	_, err = r.nvcClient.NetAppObjectStoreGet(objectStoreUUID)
	if err != nil {
		log.WithFields(log.Fields{
			"objectStore": objectStoreUUIDStr,
			"error":       err,
		}).Debug("error getting ObjectStore")
		return nil, err
	}

	if r.nvr.Spec.Source == nil {
		return nil, fmt.Errorf("source field in spec is nil, cannot process create request")
	}
	srcVolUUIDStr := r.nvr.Spec.Source.UUID
	srcVol, err := r.nvcClient.NetAppVolumeGet(srcVolUUIDStr)
	if err != nil {
		log.WithFields(log.Fields{
			"volume": srcVolUUIDStr,
			"error":  err,
		}).Debug("error getting source volume")
		return nil, err
	}

	// Check if source volume belongs to the storagepool supplied in the replication cr
	if srcVol.Labels[quarkconfig.StoragePoolLabel] != storagePoolUUIDStr {
		return nil, fmt.Errorf("source volume StoragePool (%v) not the same as StoragePool supplied in custom resource (%v)", srcVol.Labels[quarkconfig.StoragePoolLabel], storagePoolUUIDStr)
	}
	// check if source volume is already used in a backup/restore volume replication CQUARK-4870
	volExists, err := r.nvcClient.isVolumeUsedInVolumeReplication(srcVolUUIDStr)
	if err != nil {
		return nil, fmt.Errorf("error checking if volume %v is used in volume replication: %v", srcVolUUIDStr, err)
	} else if volExists {
		return nil, fmt.Errorf("source volume %v is already used in a volume replication", srcVolUUIDStr)
	}

	// check if snapshot exists for the volume CQUARK-4850
	snapshotExists, err := r.nvcClient.snapShotExistsForVolume(srcVolUUIDStr, r.nvr.Spec.SnapshotToTransfer)
	if err != nil {
		log.WithFields(log.Fields{
			"snapshotToTransfer": r.nvr.Spec.SnapshotToTransfer,
			"volume":             srcVolUUIDStr,
			"error":              err,
		}).Debug("error checking if snapshot exists for volume")
		return nil, err
	} else if !snapshotExists {
		log.WithFields(log.Fields{
			"snapshotToTransfer": r.nvr.Spec.SnapshotToTransfer,
			"volume":             srcVolUUIDStr,
			"error":              err,
		}).Debug("snapshot does not exist for volume")
		return nil, errors.NewNotFound(v1alpha1.Resource("NetAppVolumeSnapshot"), r.nvr.Spec.SnapshotToTransfer)
	}

	ownerID := r.nvr.Labels[quarkconfig.OwnerIDLabel]
	storagePoolUUID, _ := uuid.FromString(storagePoolUUIDStr)
	volReplUUID, _ := uuid.FromString(r.nvr.Name)
	srcVolUUID, _ := uuid.FromString(srcVolUUIDStr)
	destUUID, _ := uuid.FromString(r.nvr.Spec.Destination.UUID)

	// check if snapshot is already backed up for the volume CQUARK-5103
	snapExists, err := r.snapshotExistInCloudBackupList(r.nvr.Spec.SnapshotToTransfer, storagePoolUUID, srcVolUUID)
	// If there was a problem in getting the snapshot from cloudBackupList, return error.
	// If cloud backup list is not found, that is fine because there are no backups for this storage pool and we continue.
	if err != nil && !errors.IsNotFound(err) {
		return nil, err
	}
	if snapExists {
		return nil, fmt.Errorf("snapshot %v already backed up for volume %v", r.nvr.Spec.SnapshotToTransfer, srcVolUUID)
	}

	cloudbackupList, err := constructNetAppVolumeCloudBackupList(
		ownerID,
		destUUID,
		storagePoolUUID,
		srcVolUUID,
		objectStoreUUID,
		volReplUUID,
	)
	if err != nil {
		return nil, err
	}

	nvr, err := r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Create(
		context.TODO(), r.nvr, metav1.CreateOptions{})
	if err != nil {
		// If already created in a previous attempt, don't error out and continue with operations.
		if !errors.IsAlreadyExists(err) {
			return nil, err
		}
		log.WithFields(log.Fields{
			"sourceVolume":      r.nvr.Spec.Source.UUID,
			"volumeReplication": r.nvr.Name,
		}).Debug("NetAppVolumeReplication (backup) CR already exists")
	}

	// cloudbackupList will have owner ref set to the nvr created above
	cloudbackupList, err = r.nvcClient.netAppVolumeCloudBackupListCreate(cloudbackupList)
	if err != nil {
		// If already exists, return without error, idempotent behavior of this API.
		if errors.IsAlreadyExists(err) {
			log.WithFields(log.Fields{
				"sourceVolume":      r.nvr.Spec.Source.UUID,
				"volumeReplication": r.nvr.Name,
			}).Debug("NetAppVolumeCloudBackupList for volume already exists")
			return nvr, nil
		}
		log.Errorf("failed to create CloudBackupList %v, attempting to delete VolumeReplication %v which was already created",
			cloudbackupList.Name, nvr.Name)
		// try and delete nvr created above but not the end of world if it fails. Note that its going to be
		// very rare that api server went down immediately after successfully creating nvr in the previous step. In case
		// it went down, the delete of nvr will fail as well but as a good citizen we can try.
		delNVRErr := r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
			context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
		if delNVRErr != nil {
			log.Errorf("deletion of VolumeReplication %v failed, err: %v", nvr.Name, delNVRErr)
		}
		return nil, err
	}
	return nvr, nil
}

func (r *nvrBackup) Delete() error {
	if snapshotTransferInProgress(r.nvr) {
		return fmt.Errorf("backup in progress, rejecting NetAppVolumeReplication delete request for %v", r.nvr.Name)
	}
	return r.nvcClient.quarkClient.QuarkV1alpha1().NetAppVolumeReplications(r.nvcClient.nvcNamespace).Delete(
		context.TODO(), r.nvr.Name, metav1.DeleteOptions{})
}

func (r *nvrBackup) Initialize(transferSnapshot string) error {
	// For backup, Create operation sets the spec.Mirrorstate to Mirrored so no need
	// for initialize operation
	return fmt.Errorf("unsupported operation for backup")
}

func (r *nvrBackup) Resync() error {
	return fmt.Errorf("unsupported operation for backup")
}

func (r *nvrBackup) Break() error {
	return fmt.Errorf("unsupported operation for backup")
}

// TransferSnapshot API will be used by SDE to start the transfer of subsequent snapshots (2nd snapshot
// onwards) after the first snapshot has already been transferred through the Create API.
// Backup uses spec/status.SnapshotToTransfer field
func (r *nvrBackup) TransferSnapshot(snapshotName string) error {
	storagePoolUUIDStr := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	storagePoolUUID, _ := uuid.FromString(storagePoolUUIDStr)
	volumeUUIDStr := r.nvr.Spec.Source.UUID
	volumeUUID, _ := uuid.FromString(volumeUUIDStr)
	if r.nvr.Spec.SnapshotToTransfer == snapshotName {
		log.Debug("snapshot transfer request already issued, nothing to do")
		return nil
	}
	if !NetAppVolumeReplicationMirrored(r.nvr) {
		return fmt.Errorf("volume replication state must be \"%v\". Current state is \"%v\"",
			quarkv1alpha1.MirrorStateMirrored, r.nvr.Status.MirrorState)
	}
	if snapshotTransferInProgress(r.nvr) {
		return fmt.Errorf("cannot initiate a new transfer while another transfer or abort is in progress")
	}

	// check if snapshot exists for the volume CQUARK-4850
	snapshotExists, err := r.nvcClient.snapShotExistsForVolume(volumeUUIDStr, snapshotName)
	if err != nil {
		log.WithFields(log.Fields{
			"snapshotName": snapshotName,
			"volume":       volumeUUIDStr,
			"error":        err,
		}).Debug("error checking if snapshot exists for volume")
		return err
	} else if !snapshotExists {
		log.WithFields(log.Fields{
			"snapshotName": snapshotName,
			"volume":       volumeUUIDStr,
			"error":        err,
		}).Debug("snapshot does not exist for volume")
		return errors.NewNotFound(v1alpha1.Resource("NetAppVolumeSnapshot"), snapshotName)
	}

	// check if snapshot is already backed up for the volume CQUARK-5103
	snapExists, err := r.snapshotExistInCloudBackupList(snapshotName, storagePoolUUID, volumeUUID)
	if err != nil {
		return fmt.Errorf("could not verify if snpshot exist in cloudBackupList %v", err)
	}
	if snapExists {
		return fmt.Errorf("snapshot %v already backed up for volume %v", snapshotName, volumeUUIDStr)
	}

	volumeReplicationCopy := r.nvr.DeepCopy()
	volumeReplicationCopy.Spec.SnapshotToTransfer = snapshotName
	_, err = r.nvcClient.NetAppVolumeReplicationUpdate(volumeReplicationCopy)
	return err
}

func (r *nvrBackup) TransferAbort() error {
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateQuiesced {
		return nil
	}
	r.nvr.Spec.AbortTransferCounter += 1
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateQuiesced

	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}

// ScheduleModify is not needed in initial release - global scheduler is handling it
func (r *nvrBackup) ScheduleModify(schedule quarkv1alpha1.NetAppVolumeReplicationScheduleType) error {
	return fmt.Errorf("unsupported operation for backup")
}

func (r *nvrBackup) TransferResume(snapshotName string) error {
	storagePoolUUIDStr := r.nvr.Labels[quarkconfig.StoragePoolLabel]
	storagePoolUUID, _ := uuid.FromString(storagePoolUUIDStr)
	volumeUUIDStr := r.nvr.Spec.Source.UUID
	volumeUUID, _ := uuid.FromString(volumeUUIDStr)
	// r.nvr.Spec.MirrorState is quarkv1alpha1.MirrorStateMirrored, resume already issued, no-op here
	if r.nvr.Spec.MirrorState == quarkv1alpha1.MirrorStateMirrored {
		return nil
	}
	if r.nvr.Spec.MirrorState != quarkv1alpha1.MirrorStateQuiesced {
		return fmt.Errorf("current mirrorState \"%v\" is not \"%v\", nothing to resume", r.nvr.Spec.MirrorState, quarkv1alpha1.MirrorStateQuiesced)
	}
	r.nvr.Spec.MirrorState = quarkv1alpha1.MirrorStateMirrored
	if snapshotName != "" {
		// check if snapshot exists for the volume
		snapshotExists, err := r.nvcClient.snapShotExistsForVolume(r.nvr.Spec.Source.UUID, snapshotName)
		if err != nil {
			log.WithFields(log.Fields{
				"snapshotName": snapshotName,
				"volume":       r.nvr.Spec.Source.UUID,
				"error":        err,
			}).Debug("error checking if snapshot exists for volume")
			return err
		} else if !snapshotExists {
			log.WithFields(log.Fields{
				"snapshotName": snapshotName,
				"volume":       r.nvr.Spec.Source.UUID,
				"error":        err,
			}).Debug("snapshot does not exist for volume")
			return errors.NewNotFound(v1alpha1.Resource("NetAppVolumeSnapshot"), snapshotName)
		}

		// check if snapshot is already backed up for the volume CQUARK-5103
		snapExists, err := r.snapshotExistInCloudBackupList(snapshotName, storagePoolUUID, volumeUUID)
		if err != nil {
			return fmt.Errorf("could not verify if snpshot exist in cloudBackupList %v", err)
		}
		if snapExists {
			return fmt.Errorf("snapshot %v already backed up for volume %v", snapshotName, volumeUUIDStr)
		}

		r.nvr.Spec.SnapshotToTransfer = snapshotName
	}
	_, err := r.nvcClient.NetAppVolumeReplicationUpdate(r.nvr)
	return err
}

func (r *nvrBackup) snapshotExistInCloudBackupList(snapshotName string, storagePoolUUID, volumeUUID uuid.UUID) (bool, error) {
	cloudBackupList, err := r.nvcClient.NetAppVolumeCloudBackupListGetForSpAndVolume(storagePoolUUID, volumeUUID)
	if err != nil {
		return false, err
	}
	for _, backup := range cloudBackupList.Status.Backups {
		if backup.Name == snapshotName {
			return true, nil
		}
	}
	return false, nil
}
