package util

import (
	"reflect"

	"quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// IsFeatureFlagEnabled returns true if the feature flag is enabled.
func IsFeatureFlagEnabled(featureFlags v1alpha1.FeatureFlags, featureFlagName v1alpha1.FeatureFlagName) bool {
	flag, ok := featureFlags[featureFlagName]
	return ok && flag.State == v1alpha1.Enabled
}

// IsFeatureFlagEnabledInQuarkSpec returns true if the feature flag is enabled in the Quark spec.
func IsFeatureFlagEnabledInQuarkSpec(quark *v1alpha1.Quark, featureFlagName v1alpha1.FeatureFlagName) bool {
	return IsFeatureFlagEnabled(quark.Spec.FeatureFlags, featureFlagName)
}

// IsFeatureFlagEnabledInQuarkStatus returns true if the feature flag is enabled in the Quark status.
func IsFeatureFlagEnabledInQuarkStatus(quark *v1alpha1.Quark, featureFlagName v1alpha1.FeatureFlagName) bool {
	return IsFeatureFlagEnabled(quark.Status.FeatureFlags, featureFlagName)
}

// IsStoragePoolFeatureFlagEnabled returns true if the storage pool feature flag is enabled.
func IsStoragePoolFeatureFlagEnabled(featureFlags v1alpha1.StoragePoolFeatureFlags, featureFlagName v1alpha1.FeatureFlagName) bool {
	return featureFlags[featureFlagName] == v1alpha1.Enabled
}

// IsFeatureFlagEnabledInStoragePoolSpec returns true if the feature flag is enabled in the storage pool spec.
func IsFeatureFlagEnabledInStoragePoolSpec(storagePool *v1alpha1.NetAppStoragePool, featureFlagName v1alpha1.FeatureFlagName) bool {
	return IsStoragePoolFeatureFlagEnabled(storagePool.Spec.FeatureFlags, featureFlagName)
}

// IsFeatureFlagEnabledInStoragePoolStatus returns true if the feature flag is enabled in the storage pool status.
func IsFeatureFlagEnabledInStoragePoolStatus(storagePool *v1alpha1.NetAppStoragePool, featureFlagName v1alpha1.FeatureFlagName) bool {
	return IsStoragePoolFeatureFlagEnabled(storagePool.Status.FeatureFlags, featureFlagName)
}

// CompareFeatureFlags checks if two sets of feature flags are equal.
func CompareFeatureFlags(new, old v1alpha1.FeatureFlags) bool {
	return reflect.DeepEqual(new, old)
}

// CompareStoragePoolFeatureFlags checks if two sets of storagepool feature flags are equal.
func CompareStoragePoolFeatureFlags(new, old v1alpha1.StoragePoolFeatureFlags) bool {
	return reflect.DeepEqual(new, old)
}

// IsFeatureFlagRolledback returns true if the feature flag has been rolled back in quark.
func IsFeatureFlagRolledback(quark *v1alpha1.Quark, featureFlagName v1alpha1.FeatureFlagName) bool {
	oldFlag, oldOk := quark.Status.FeatureFlags[featureFlagName]
	newFlag, newOk := quark.Spec.FeatureFlags[featureFlagName]
	return oldOk && newOk && oldFlag.State == v1alpha1.Enabled && newFlag.State == v1alpha1.Disabled
}

// IsStoragePoolFeatureFlagRolledback returns true if the feature flag has been rolled back in the storage pool.
func IsStoragePoolFeatureFlagRolledback(storagePool *v1alpha1.NetAppStoragePool, featureFlagName v1alpha1.FeatureFlagName) bool {
	return storagePool.Status.FeatureFlags[featureFlagName] == v1alpha1.Enabled && storagePool.Spec.FeatureFlags[featureFlagName] == v1alpha1.Disabled
}

// GetStoragePoolFeatureFlagsFromQuark returns the storage pool scoped feature flags from quark.
func GetStoragePoolFeatureFlagsFromQuark(quark *v1alpha1.Quark) v1alpha1.StoragePoolFeatureFlags {
	storagePoolFeatureFlags := make(v1alpha1.StoragePoolFeatureFlags)
	for flag, value := range quark.Spec.FeatureFlags {
		if value.StoragePoolType != "" {
			storagePoolFeatureFlags[flag] = value.State
		}
	}
	return storagePoolFeatureFlags
}

// GetUpdatedStoragePoolFeatureFlags returns the updated storage pool feature flags based on the provided Quark and storage pool.
// It also returns the feature flags that failed validation.
func GetUpdatedStoragePoolFeatureFlags(quark *v1alpha1.Quark, storagePool *v1alpha1.NetAppStoragePool) (updatedFlags, failedValidationFlags v1alpha1.StoragePoolFeatureFlags) {
	failedValidation := make(v1alpha1.StoragePoolFeatureFlags)
	newStoragePoolFeatureFlags := GetStoragePoolFeatureFlagsFromQuark(quark)
	existingStoragePoolFeatureFlags := storagePool.Spec.FeatureFlags

	for flagName, newFlagState := range newStoragePoolFeatureFlags {
		if existingFlag, ok := existingStoragePoolFeatureFlags[flagName]; ok {
			switch quark.Spec.FeatureFlags[flagName].StoragePoolType {
			case v1alpha1.NonReversibleOnceEnabled:
				// If the feature flag is non reversible after it is enabled it cannot be disabled. So,
				// if the flag is enabled it'll be kept enabled irrespective of the value in the new feature flag.
				if existingFlag == v1alpha1.Enabled && newFlagState == v1alpha1.Disabled {
					failedValidation[flagName] = newStoragePoolFeatureFlags[flagName]
					newStoragePoolFeatureFlags[flagName] = existingFlag
				}
			case v1alpha1.NonReversible:
				// If the feature flag is non reversible it cannot be toggled once it's set.
				if existingFlag != newFlagState {
					failedValidation[flagName] = newStoragePoolFeatureFlags[flagName]
					newStoragePoolFeatureFlags[flagName] = existingFlag
				}
			}
		}
	}

	return newStoragePoolFeatureFlags, failedValidation
}
