package helm

import (
	"context"
	"fmt"
	"os"
	"time"

	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	"quark.netapp.io/volume-controller/pkg/log"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	apierrors "k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/meta"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	"k8s.io/client-go/discovery"
	"k8s.io/client-go/discovery/cached/memory"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/restmapper"
	"k8s.io/client-go/tools/clientcmd"

	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
	"quark.netapp.io/volume-controller/pkg/util"
)

// migrateQuarkResourcesToHelm takes a kubeconfig, quark name and quark namespace and adds annotations to existing quark resources
// to allow quark to be upgraded with the helm workflow
func (h *HelmReconciler) migrateQuarkResourcesToHelm() error {
	log.Debugln("Migrating resources to helm")
	quarkReleaseName := "quark"
	quarkOperatorReleaseName := "quark-operator"

	resources := []struct {
		Resource         string
		Name             string
		Namespace        string
		ReleaseName      string
		ReleaseNamespace string
	}{
		// For Quark, Operator Deployment, Operator Service Account, we have the namespace as atom-quark which is the
		// same as the helm release-namespace. For Kube state metrics Service Account the namespace is quark-system.
		// Whereas for resources like ClusterRole, ClusterRoleBinding and CRD namespace does not matter hence
		// namespace is passed as empty.
		{ResourceQuark, GetQuarkCRName(h.Values, h.Namespace), h.Namespace, quarkReleaseName, h.Namespace},
		{ResourceDeployment, quarkconfig.QuarkOperatorDeploymentName, h.Namespace, quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRoleBinding, quarkconfig.QuarkOperatorClusterRoleBinding, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRoleBinding, quarkconfig.KubeStateMetricsClusterRoleBinding, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRoleBinding, quarkconfig.NodeVolClusterRoleBinding, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRole, quarkconfig.QuarkOperatorClusterRole, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRole, quarkconfig.NodeVolClusterRole, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceRoleBinding, quarkconfig.NodeVolRoleBinding, quarkconfig.QuarkSystemNamespace, quarkOperatorReleaseName, h.Namespace},
		{ResourceRole, quarkconfig.NodeVolRole, quarkconfig.QuarkSystemNamespace, quarkOperatorReleaseName, h.Namespace},
		{ResourceClusterRole, quarkconfig.KubeStateMetricsClusterRole, "", quarkOperatorReleaseName, h.Namespace},
		{ResourceServiceAccount, quarkconfig.QuarkOperatorSA, h.Namespace, quarkOperatorReleaseName, h.Namespace},
		{ResourceServiceAccount, quarkconfig.KubeStateMetricsSA, quarkconfig.QuarkSystemNamespace, quarkOperatorReleaseName, h.Namespace},
		{ResourceCrd, quarkconfig.QuarkCRDName, "", quarkReleaseName, h.Namespace},
	}

	for _, r := range resources {
		err := AnnotateAndLabelK8sResource(h.Kubeconfig, r.Resource, r.Name, r.Namespace, r.ReleaseName, r.ReleaseNamespace)
		if err != nil {
			if apierrors.IsNotFound(err) {
				log.Debugf("Could not find resource %s, skipping", r.Name)
			} else {
				return err
			}
		}
	}
	return nil
}

// AnnotateAndLabelK8sResource annotates an existing k8s resource with a specific helm release name, so helm is aware of it for upgrade/install
func AnnotateAndLabelK8sResource(config *rest.Config, resource, name, namespace, releaseName, releaseNamespace string) error {
	log.WithFields(log.Fields{
		"resource":  resource,
		"name":      name,
		"namespace": namespace,
	}).Debugln("Running AnnotateAndLabelK8sResource")

	quarkClient, err := quarkclientset.NewForConfig(config)
	if err != nil {
		return err
	}

	apiextensionsclientset, err := apiextensionsclient.NewForConfig(config)
	if err != nil {
		return err
	}

	clientset, err := kubernetes.NewForConfig(config)
	if err != nil {
		return err
	}

	return annotateAndLabelK8sResource(quarkClient, apiextensionsclientset, clientset, resource, name, namespace, releaseName, releaseNamespace)
}

// annotateAndLabelK8sResource annotates an existing k8s resource with a specific helm release name, so helm is aware of it for upgrade/install
func annotateAndLabelK8sResource(quarkClient quarkclientset.Interface, apiExClient apiextensionsclient.Interface, kubeClientset kubernetes.Interface, resource, name, namespace, releaseName, releaseNamespace string) error {
	ctx := context.Background()
	var updateFunc func(context.Context, metav1.UpdateOptions) (metav1.Object, error)

	switch resource {
	case ResourceDeployment:

		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			deployment, err := kubeClientset.AppsV1().Deployments(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(deployment, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.AppsV1().Deployments(namespace).Update(ctx, deployment, opts)
			}
			return deployment, nil
		}
	case ResourceClusterRoleBinding:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			clusterRoleBinding, err := kubeClientset.RbacV1().ClusterRoleBindings().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(clusterRoleBinding, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.RbacV1().ClusterRoleBindings().Update(ctx, clusterRoleBinding, opts)
			}
			return clusterRoleBinding, nil
		}
	case ResourceClusterRole:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			clusterRole, err := kubeClientset.RbacV1().ClusterRoles().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(clusterRole, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.RbacV1().ClusterRoles().Update(ctx, clusterRole, opts)
			}
			return clusterRole, nil
		}
	case ResourceRoleBinding:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			roleBinding, err := kubeClientset.RbacV1().RoleBindings(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(roleBinding, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.RbacV1().RoleBindings(namespace).Update(ctx, roleBinding, opts)
			}
			return roleBinding, nil
		}
	case ResourceRole:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			role, err := kubeClientset.RbacV1().Roles(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(role, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.RbacV1().Roles(namespace).Update(ctx, role, opts)
			}
			return role, nil
		}
	case ResourceServiceAccount:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			serviceAccount, err := kubeClientset.CoreV1().ServiceAccounts(namespace).Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(serviceAccount, releaseName, releaseNamespace)
			if shouldUpdate {
				return kubeClientset.CoreV1().ServiceAccounts(namespace).Update(ctx, serviceAccount, opts)
			}
			return serviceAccount, nil
		}
	case ResourceCrd:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			customResourceDefinition, err := apiExClient.ApiextensionsV1().CustomResourceDefinitions().Get(ctx, name, metav1.GetOptions{})
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(customResourceDefinition, releaseName, releaseNamespace)
			if shouldUpdate {
				return apiExClient.ApiextensionsV1().CustomResourceDefinitions().Update(ctx, customResourceDefinition, opts)
			}
			return customResourceDefinition, nil
		}
	case ResourceQuark:
		updateFunc = func(ctx context.Context, opts metav1.UpdateOptions) (metav1.Object, error) {
			quark, err := util.GetFirstQuark(quarkClient)
			if err != nil {
				return nil, err
			}
			shouldUpdate := applyAnnotationsAndLabels(quark, releaseName, releaseNamespace)
			if shouldUpdate {
				return quarkClient.QuarkV1alpha1().Quarks(quark.Spec.Namespace).Update(
					context.TODO(), quark, metav1.UpdateOptions{})
			}
			return quark, nil
		}
	default:
		return fmt.Errorf("unsupported resource type: %s", resource)
	}

	retryCount := 0
	for retryCount < 3 {
		_, err := updateFunc(ctx, metav1.UpdateOptions{})
		if err != nil {
			if apierrors.IsConflict(err) {
				log.Debugf("Resource conflict error, retrying (%d/3)...\n", retryCount+1)
				time.Sleep(2 * time.Second)
				retryCount++
			} else {
				return err
			}
		} else {
			break
		}
	}
	return nil
}

// applyAnnotationsAndLabels is a helper function that generically applies helm labels and annotations to a k8s resource
func applyAnnotationsAndLabels(obj metav1.Object, releaseName, releaseNamespace string) (needsUpdate bool) {
	labels := obj.GetLabels()
	if labels == nil {
		labels = make(map[string]string)
	}
	if val, ok := labels["app.kubernetes.io/managed-by"]; ok && val == "Helm" {
		// already managed by helm
		return false
	}
	labels["app.kubernetes.io/managed-by"] = "Helm"
	obj.SetLabels(labels)

	annotations := obj.GetAnnotations()
	if annotations == nil {
		annotations = make(map[string]string)
	}
	annotations["meta.helm.sh/release-name"] = releaseName
	annotations["meta.helm.sh/release-namespace"] = releaseNamespace
	obj.SetAnnotations(annotations)
	return true
}

// CustomRESTClientGetter implements a clientcmd k8s network interface to convert a rest.Config to a connection endpoint for helm
type CustomRESTClientGetter struct {
	config     *rest.Config
	CacheDir   *string
	KubeConfig *string

	// config flags
	ClusterName        *string
	AuthInfoName       *string
	Context            *string
	Namespace          *string
	APIServer          *string
	TLSServerName      *string
	Insecure           *bool
	CertFile           *string
	CertData           *[]byte
	KeyFile            *string
	KeyData            *[]byte
	CAFile             *string
	CAData             *[]byte
	BearerToken        *string
	Impersonate        *string
	ImpersonateUID     *string
	ImpersonateGroup   *[]string
	Username           *string
	Password           *string
	Timeout            *string
	DisableCompression *bool
	// If non-nil, wrap config f
}

func NewCustomRESTClientGetter(c *rest.Config, namespace string) *CustomRESTClientGetter {
	tru := true

	customRest := &CustomRESTClientGetter{}
	customRest.config = c
	if c.CAFile != "" {
		customRest.CAFile = &c.CAFile
	}
	if c.CAData != nil && len(c.CAData) > 0 {
		customRest.CAData = &c.CAData
	}
	if c.CertFile != "" {
		customRest.CertFile = &c.CertFile
	}
	if c.CertData != nil && len(c.CertData) > 0 {
		customRest.CertData = &c.CertData
	}
	if c.KeyFile != "" {
		customRest.KeyFile = &c.KeyFile
	}
	if c.KeyData != nil && len(c.KeyData) > 0 {
		customRest.KeyData = &c.KeyData
	}
	if c.Host != "" {
		customRest.APIServer = &c.Host
	}
	if c.Insecure {
		customRest.Insecure = &tru
	}
	if c.TLSClientConfig.ServerName != "" {
		customRest.TLSServerName = &c.ServerName
	}
	if c.BearerToken != "" {
		customRest.BearerToken = &c.BearerToken
	}
	if c.Impersonate.UserName != "" {
		customRest.Impersonate = &c.Impersonate.UserName
	}
	if c.Impersonate.UID != "" {
		customRest.ImpersonateUID = &c.Impersonate.UID
	}
	if c.Impersonate.Groups != nil && len(c.Impersonate.Groups) > 0 {
		customRest.ImpersonateGroup = &c.Impersonate.Groups
	}
	if c.Username != "" {
		customRest.Username = &c.Username
	}
	if c.Password != "" {
		customRest.Password = &c.Password
	}
	if c.Timeout.Seconds() > 0 {
		timeoutStr := c.Timeout.String()
		customRest.Timeout = &timeoutStr
	}
	return customRest
}

func (c *CustomRESTClientGetter) ToRESTConfig() (*rest.Config, error) {
	return c.config, nil
}

func (f *CustomRESTClientGetter) ToRawKubeConfigLoader() clientcmd.ClientConfig {
	loadingRules := clientcmd.NewDefaultClientConfigLoadingRules()
	// use the standard defaults for this client command
	// DEPRECATED: remove and replace with something more accurate
	loadingRules.DefaultClientConfig = &clientcmd.DefaultClientConfig

	if f.KubeConfig != nil {
		loadingRules.ExplicitPath = *f.KubeConfig
	}

	overrides := &clientcmd.ConfigOverrides{ClusterDefaults: clientcmd.ClusterDefaults}
	if f.APIServer != nil {
		overrides.ClusterInfo.Server = *f.APIServer
	}
	// bind auth info flag values to overrides
	if f.CertFile != nil {
		overrides.AuthInfo.ClientCertificate = *f.CertFile
	}
	if f.CertData != nil {
		overrides.AuthInfo.ClientCertificateData = *f.CertData
	}
	if f.KeyFile != nil {
		overrides.AuthInfo.ClientKey = *f.KeyFile
	}
	if f.KeyData != nil {
		overrides.AuthInfo.ClientKeyData = *f.KeyData
	}
	if f.BearerToken != nil {
		overrides.AuthInfo.Token = *f.BearerToken
	}
	if f.Impersonate != nil {
		overrides.AuthInfo.Impersonate = *f.Impersonate
	}
	if f.ImpersonateUID != nil {
		overrides.AuthInfo.ImpersonateUID = *f.ImpersonateUID
	}
	if f.ImpersonateGroup != nil {
		overrides.AuthInfo.ImpersonateGroups = *f.ImpersonateGroup
	}
	if f.Username != nil {
		overrides.AuthInfo.Username = *f.Username
	}
	if f.Password != nil {
		overrides.AuthInfo.Password = *f.Password
	}

	// bind cluster flags
	if f.APIServer != nil {
		overrides.ClusterInfo.Server = *f.APIServer
	}
	if f.TLSServerName != nil {
		overrides.ClusterInfo.TLSServerName = *f.TLSServerName
	}
	if f.CAFile != nil {
		overrides.ClusterInfo.CertificateAuthority = *f.CAFile
	}
	if f.CAData != nil {
		overrides.ClusterInfo.CertificateAuthorityData = *f.CAData
	}
	if f.Insecure != nil {
		overrides.ClusterInfo.InsecureSkipTLSVerify = *f.Insecure
	}
	if f.DisableCompression != nil {
		overrides.ClusterInfo.DisableCompression = *f.DisableCompression
	}

	// bind context flags
	if f.Context != nil {
		overrides.CurrentContext = *f.Context
	}
	if f.ClusterName != nil {
		overrides.Context.Cluster = *f.ClusterName
	}
	if f.AuthInfoName != nil {
		overrides.Context.AuthInfo = *f.AuthInfoName
	}
	if f.Namespace != nil {
		overrides.Context.Namespace = *f.Namespace
	}

	if f.Timeout != nil {
		overrides.Timeout = *f.Timeout
	}

	// we only have an interactive prompt when a password is allowed
	if f.Password == nil {
		return clientcmd.NewNonInteractiveDeferredLoadingClientConfig(loadingRules, overrides)
	}
	return clientcmd.NewInteractiveDeferredLoadingClientConfig(loadingRules, overrides, os.Stdin)
}

func (c *CustomRESTClientGetter) ToDiscoveryClient() (discovery.CachedDiscoveryInterface, error) {
	discoveryClient, err := discovery.NewDiscoveryClientForConfig(c.config)
	if err != nil {
		return nil, err
	}
	return memory.NewMemCacheClient(discoveryClient), nil
}

func (c *CustomRESTClientGetter) ToRESTMapper() (meta.RESTMapper, error) {
	discoveryClient, err := c.ToDiscoveryClient()
	if err != nil {
		return nil, err
	}

	mapper := restmapper.NewDeferredDiscoveryRESTMapper(discoveryClient)
	expander := restmapper.NewShortcutExpander(mapper, discoveryClient, nil)
	return expander, nil
}
