package util

import (
	"fmt"
	"strings"
	"time"

	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/tools/record"
	"quark.netapp.io/volume-controller/pkg/log"

	// Quark
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

const SubsystemInternalErrorPrifix = "Internal: "

// RecordStoragePoolSubsystemsEvent is used to record storagepool's subsystems event on server
func RecordStoragePoolSubsystemsEvent(storagePool *quarkv1alpha1.NetAppStoragePool, recorder record.EventRecorder) {
	if storagePool.Status.GranularHealth != nil {
		for subsystem, subsystemHealth := range storagePool.Status.GranularHealth {
			if event := SetStoragepoolSubsystemEvent(subsystem, subsystemHealth.Message, subsystemHealth.Status); event != nil {
				recorder.Event(storagePool, event.Type, event.Reason, event.Message)
			}
		}
	}
}

func SetStoragepoolSubsystemEvent(module, msg string, status corev1.ConditionStatus) *corev1.Event {
	// Avoiding boot subsystem events as we're already raising online condition events.
	if module == quarkconfig.BootSubsystem {
		return nil
	}

	var (
		eventReason = quarkconfig.NetAppStoragePoolSubsystemStatusChanged
		eventType   = corev1.EventTypeNormal
		eventState  = quarkconfig.SubsystemReady
		eventMsg    string
	)
	if status == corev1.ConditionFalse {
		eventType, eventState = corev1.EventTypeWarning, quarkconfig.SubsystemNotReady
	}
	if msg != "" {
		eventMsg = fmt.Sprintf("%s subsystem is in %s state with message %s", module, eventState, TrimInternalSubsystemError(msg))
	} else {
		eventMsg = fmt.Sprintf("%s subsystem is in %s state", module, eventState)
	}

	return &corev1.Event{
		Type:    eventType,
		Reason:  eventReason,
		Message: eventMsg,
	}
}

// returns true if storage pool online condition exists
func IsStoragePoolHealthInitialized(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return getCondition(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolOnline) != nil
}

// returns true if granular health is enabled and is initialized in storage pool
func UseStoragePoolGranularHealth(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return storagePool.Status.GranularHealth != nil
}

// returns storagepool online from online condition(if present)
func IsNetAppStoragePoolOnline(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	onlineCondition := getCondition(storagePool.Status.Conditions, quarkv1alpha1.NetAppStoragePoolOnline)

	return onlineCondition != nil && onlineCondition.Status == corev1.ConditionTrue
}

// Mark storagePool as offline if it is already online. When pod restarts, podrick will mark storagePool as online
func MakeNetAppStoragePoolOffline(storagePool *quarkv1alpha1.NetAppStoragePool, msg string) *quarkv1alpha1.NetAppStoragePool {
	conditions := storagePool.Status.Conditions

	currTime := metav1.NewTime(time.Now())
	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.NetAppStoragePoolOnline && conditions[i].Status == corev1.ConditionTrue {
			conditions[i].LastHeartbeatTime = currTime
			conditions[i].LastTransitionTime = currTime
			conditions[i].Status = corev1.ConditionFalse
			conditions[i].Reason = quarkconfig.NetAppStoragePoolNotReady
			conditions[i].Message = msg
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"conditions":  conditions,
			}).Debug("Updated online condition.")
			break
		}
	}
	storagePool.Status.Conditions = conditions

	// Only if Granular health is enabled
	if storagePool.Status.GranularHealth != nil {
		granularHealth := make(quarkv1alpha1.StoragePoolGranularHealthMap)
		for subsystem := range storagePool.Status.GranularHealth {
			// All the errors causing SP marking to offline will be internal errors.
			// Hence converting to msg in to internal error.
			if !IsInternalSubsystemError(msg) {
				msg = MakeInternalSubsystemError(msg)
			}

			// MArking each subsystem as notReady.
			granularHealth[subsystem] = quarkv1alpha1.NetAppStoragePoolSubsystem{Status: corev1.ConditionFalse, Message: msg, LastTransitionTime: currTime}
		}
		storagePool.Status.GranularHealth = granularHealth
	}

	return storagePool
}

// GetVolumeSubsystems returns volume associated subsystems based on its protocol.
func GetVolumeSubsystems(storagePool *quarkv1alpha1.NetAppStoragePool, volume *quarkv1alpha1.NetAppVolume) []string {
	var subsystems []string

	if CheckNfsVolume(volume.Spec.Protocol) {
		if storagePool.Status.LdapEnabled {
			subsystems = append(subsystems, quarkconfig.NfsLdapSubsystem)
			if volume.Spec.KerberosEnabled {
				subsystems = append(subsystems, quarkconfig.NfsKerberosSubSystem)
			}
			// Can be a case where LDAP is not configured but kerberos is.
		} else if volume.Spec.KerberosEnabled {
			subsystems = append(subsystems, quarkconfig.NfsKerberosSubSystem)
		} else {
			// Default NFS status is same as BOOT subsystem status
			subsystems = append(subsystems, quarkconfig.BootSubsystem)
		}
	} else {
		subsystems = append(subsystems, quarkconfig.SmbSubsystem)
	}

	return subsystems
}

// GetVolumeSubsystemStatus returns subsystems linked to volume, consolidated online status and message.
func GetVolumeSubsystemStatus(storagePool *quarkv1alpha1.NetAppStoragePool, volume *quarkv1alpha1.NetAppVolume, granularHealth quarkv1alpha1.StoragePoolGranularHealthMap) ([]string, bool, string) {
	var subsystems = GetVolumeSubsystems(storagePool, volume)

	// If the Storagepool is Offline then all the subsystems will be in notReady state. But in this case Podrick/DMAP can be down.
	// Hence these subsystems status in Storagepool CR may not have updated to notReady and the current state could be a stale one.
	// So, setting the subsystem status as offline if Storagepool is offline
	if granularHealth != nil {
		if granularHealth[quarkconfig.BootSubsystem].Status != corev1.ConditionTrue {
			return subsystems, false, fmt.Sprintf("%s Subsystem: %s", quarkconfig.BootSubsystem, granularHealth[quarkconfig.BootSubsystem].Message)
		}
		// Get subsystem details from granular health data. Ldap subsystem in GetVolumeSubsystems would be appended prior to kerberos subsystem.
		// If LDAP subsystem is down, kerberos volumes will be marked offline. Also, it will only provide message for subsystems which are false.
		// We don't need to worry on other subysystem if its parent is down.
		for _, subsystem := range subsystems {
			if granularHealth[subsystem].Status != corev1.ConditionTrue {
				if granularHealth[subsystem].Message == "" {
					return subsystems, false, fmt.Sprintf("%s Subsystem is not ready", subsystem)
				}
				return subsystems, false, fmt.Sprintf("%s Subsystem: %s", subsystem, granularHealth[subsystem].Message)
			}
		}
	}

	return subsystems, true, ""
}

func IsInternalSubsystemError(message string) bool {
	return strings.HasPrefix(message, SubsystemInternalErrorPrifix)
}

func TrimInternalSubsystemError(message string) string {
	return strings.TrimPrefix(message, SubsystemInternalErrorPrifix)
}

func MakeInternalSubsystemError(message string) string {
	return SubsystemInternalErrorPrifix + message
}
